<?php

namespace NoccyLabs\Daemon;

use Psr\Log\LoggerTrait;

abstract class AbstractDaemon
{
    use LoggerTrait;

    /** @var bool True as long as the daemon is running, setting to false will cause shutdown */
    protected $isRunning = false;

    /** @var array Active timeouts */
    protected $activeTimeouts = [];

    /** @var array Active intervals */
    protected $activeIntervals = [];

    /** @var string Date format for the log */
    protected $dateFormat = 'M d H:i:s';

    /**
     * Set up the daemon
     *
     */
    abstract protected function onSetup();

    /**
     * Shutdown function
     *
     */
    abstract protected function onShutdown();

    /**
     * Loop function
     *
     */
    abstract protected function onLoop();

    /**
     * Signal handler
     *
     */
    public function onSignal($sigNo)
    {
        if ($sigNo == SIGINT) {
            $this->stop();
            echo "\n";
            $this->info("Received SIGINT");
        }
    }

    /**
     * Check and update timeouts and intervals
     *
     */
    protected function checkTimers()
    {
        foreach ($this->activeTimeouts as $index=>$timeout) {
            if (time() >= $timeout['next']) {
                call_user_func($timeout['callback'], $timeout['userdata']);
                unset($this->activeTimeouts[$index]);
            }
        }

        foreach ($this->activeIntervals as $index=>$interval) {
            if (time() >= $interval['next']) {
                call_user_func($interval['callback'], $interval['userdata']);
                $this->activeIntervals[$index]['next'] = time() + $interval['seconds'];
            }
        }
    }

    /**
     * Add a timeout to be called after the specified number of seconds. The timeout will
     * be removed once fired.
     *
     * @param int $seconds The number of seconds to wait before calling the callback
     * @param callable $callback Callback function
     * @param mixed|null $userdata Data to pass to the callback function
     */
    protected function addTimeout($seconds, callable $callback, $userdata=null)
    {
        $this->activeTimeouts[] = [
            "seconds" => $seconds,
            "next" => time() + $seconds,
            "callable" => $callback,
            "userdata" => $userdata,
        ];
    }

    /**
     * Register a function to be called every specified number of seconds.
     *
     * @param int $seconds The number of seconds to wait between calling the callback
     * @param callable $callback Callback function
     * @param mixed|null $userdata Data to pass to the callback function
     */
    protected function addInterval($seconds, callable $callback, $userdata=null)
    {
        $this->activeIntervals[] = [
            "seconds" => $seconds,
            "next" => time() + $seconds,
            "callback" => $callback,
            "userdata" => $userdata,
        ];
    }

    /**
     * {@inheritDoc}
     *
     */
    public function log($level, $message, array $context = array())
    {
        printf("%s [%s] %s\n", date($this->dateFormat), $level, $message);
    }

    /**
     * Start the daemon.
     *
     *
     */
    final public function run()
    {
        $this->onSetup();

        pcntl_signal(SIGINT, [ $this, "onSignal" ]);

        $this->isRunning = true;
        do {
            $this->onLoop();
            usleep(100000);
            $this->checkTimers();
            pcntl_signal_dispatch();
        } while ($this->isRunning);

        $this->onShutdown();
    }

    final public function stop()
    {
        $this->isRunning = false;
    }

    /**
     * Static entry point.
     *
     */
    public static function main()
    {
        // Create a new instance of the requested daemon class
        $daemon = get_called_class();
        $inst = new $daemon();

        // Start the daemon
        $inst->run();
    }
}
